/*
 *                           0BSD 
 * 
 *                    BSD Zero Clause License
 * 
 *  Copyright (c) 2019 Hermann Meyer
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.

 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#pragma once

#ifndef XWIDGETS_H_
#define XWIDGETS_H_


/** 
 * @file xwidgets.h
 * @brief xwidgets.h include some predefined widgets for libxputty, include this to use them 
 * if you would only use libxputty and define your own widgets, include xputt.h 
 * instead this one in your project.
 */


/**@example Makefile
 * Simple example of how to build this
 */

// widget header
#include "xbutton.h"
#include "xslider.h"
#include "xknob.h"
#include "xmenu.h"
#include "xcombobox.h"
#include "xtooltip.h"
#include "xmeter.h"
#include "xlabel.h"
#include "xframe.h"
#include "xlistbox.h"
#include "xlistview.h"
#include "xmultilistview.h"
#include "xplayhead.h"
#include "xtuner.h"
#include "xvaluedisplay.h"
#include "xwaveview.h"
#include "xtabbox.h"
#include "xdrawing_area.h"
#include "xkeyboard.h"


#endif //XWIDGETS_H_
